from pydantic import BaseModel

from vis3.internal.schema.state import State


class BucketCreateBody(BaseModel):
    name: str | None = None
    path: str
    endpoint: str
    keychain_id: int


class BucketCreatePayload(BucketCreateBody):
    created_by: int
    state: State = State.ENABLED


class BucketUpdateBody(BaseModel):
    name: str | None = None
    path: str | None = None
    endpoint: str | None = None
    keychain_id: int | None = None
    state: State | None = None

class BucketUpdatePayload(BucketUpdateBody):
    updated_by: int | None = None



import threading
import time
import webbrowser
from typing import Any

import uvicorn
from fastapi import FastAPI, Response
from fastapi.staticfiles import StaticFiles
from loguru import logger
from typer import Typer

from vis3.alembic_vis3.run_migrate import run_db_migrations
from vis3.internal.api import initial_routers
from vis3.internal.common.db import init_tables
from vis3.internal.common.exceptions import add_exception_handler
from vis3.internal.config import settings
from vis3.version import version

app = FastAPI(
  title="Vis3",
  description="Visualize s3 data",
  version=version,
  terms_of_service="",
  contact={
      "name": "Vis3",
      "url": "https://github.com/OpenDataLab/Vis3",
      "email": "shenguanlin@pjlab.org.cn",
  },
  license_info={
      "name": "Apache 2.0",
      "url": "https://www.apache.org/licenses/LICENSE-2.0.html",
  },
)


class NoCacheStaticFiles(StaticFiles):
    def __init__(self, *args: Any, **kwargs: Any):
        self.cachecontrol = "max-age=0, no-cache, no-store, must-revalidate"
        self.pragma = "no-cache"
        self.expires = "0"
        super().__init__(*args, **kwargs)

    def file_response(self, *args: Any, **kwargs: Any) -> Response:
        resp = super().file_response(*args, **kwargs)
        
        # No cache for html files
        if resp.media_type == "text/html":
            resp.headers.setdefault("Cache-Control", self.cachecontrol)
            resp.headers.setdefault("Pragma", self.pragma)
            resp.headers.setdefault("Expires", self.expires)
            
        return resp


init_tables()
run_db_migrations()
initial_routers(app)
add_exception_handler(app)

app.mount("", NoCacheStaticFiles(packages=["vis3.internal"], html=True))

cli = Typer()

@cli.callback(invoke_without_command=True)
def main(
    host: str = "localhost", 
    port: int = 8000,
    open: bool = False
):
    if port:
        settings.PORT = str(port)  # 确保PORT是字符串
    if host:
        settings.HOST = host
    
    if open:
        def open_browser():
            time.sleep(2)
            url = f"http://{settings.HOST}:{settings.PORT}"
            logger.info(f"Opening browser: {url}")
            webbrowser.open(url)
        
        # 在后台线程中打开浏览器
        browser_thread = threading.Thread(target=open_browser)
        browser_thread.daemon = True
        browser_thread.start()
        
    logger.info(f"Start server: http://{settings.HOST}:{settings.PORT}")
    uvicorn.run(app=app, host=settings.HOST, port=int(settings.PORT))
        

if __name__ == "__main__":
    cli()


from pydantic import BaseModel
from vis3.internal.schema.state import State


class KeychainCreateBody(BaseModel):
    name: str
    access_key_id: str
    secret_key_id: str


class KeychainCreatePayload(KeychainCreateBody):
    created_by: int
    state: State = State.ENABLED


class KeychainUpdateBody(BaseModel):
    name: str | None = None
    access_key_id: str | None = None
    secret_key_id: str | None = None
    state: State | None = None

class KeychainUpdatePayload(KeychainUpdateBody):
    updated_by: int



from typing import Generator

from loguru import logger
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from vis3.internal.config import settings

engine = None
database_url = settings.DATABASE_URL


# connect_args is needed only for SQLite. It's not needed for other databases
engine = create_engine(
    database_url,
    connect_args={"check_same_thread": False},
    echo=True,
)

SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)

Base = declarative_base()

# create database tables
def init_tables() -> None:
    Base.metadata.create_all(bind=engine)

def get_db() -> Generator:
    db = None
    try:
        db = SessionLocal()
        yield db
        db.commit()
    except Exception as e:
        logger.error(e)
        if db:
            db.rollback()

        raise e
    finally:
        if db:
            db.close()



"""remove_name_in_bucket

Revision ID: 36abca001cd5
Revises: 
Create Date: 2025-06-10 15:18:38.084648

"""
from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = '36abca001cd5'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    tables = inspector.get_table_names()
    # 删除 bucket 表中的 name 列
    if 'bucket' in tables and 'name' in inspector.get_columns('bucket'):
        op.drop_column('bucket', 'name')
        op.execute("UPDATE bucket SET name = NULL")


def downgrade() -> None:
    """Downgrade schema."""
    pass



from typing import List, Union

from fastapi import APIRouter, Depends, Request, status
from sqlalchemy.orm import Session

from vis3.internal.api.dependencies.auth import get_auth_user_or_error
from vis3.internal.api.v1.schema.request.bucket import (
    BucketCreateBody,
    BucketUpdateBody,
    BucketUpdatePayload,
)
from vis3.internal.api.v1.schema.response import ItemResponse, ListResponse, OkResponse
from vis3.internal.api.v1.schema.response.bucket import BucketResponse, PathType
from vis3.internal.common.db import get_db
from vis3.internal.common.exceptions import AppEx, ErrorCode
from vis3.internal.config import settings
from vis3.internal.crud.bucket import bucket_crud
from vis3.internal.crud.keychain import keychain_crud
from vis3.internal.models.user import User
from vis3.internal.service.bucket import (
    get_bucket,
    get_buckets_or_objects,
    preview_file,
)
from vis3.internal.utils import ping_host, validate_path_accessibility
from vis3.internal.utils.path import is_s3_path, split_s3_path

router = APIRouter(tags=["buckets"])


@router.get(
    "/bucket",
    summary="获取所有 bucket 列表",
    response_model=Union[ListResponse[BucketResponse], ItemResponse[BucketResponse]],
)
async def read_bucket_request(
    request: Request,
    id: int | None = None,
    path: str = None,
    page_no: int = 1,
    page_size: int = 10,
    db: Session = Depends(get_db),
    current_user: User | None = Depends(get_auth_user_or_error),
):
    """
    获取指定 bucket 下的所有对象
    """
    result = await get_buckets_or_objects(
        path=path,
        id=id,
        page_no=page_no,
        page_size=page_size,
        db=db,
        user_id=current_user.id if current_user else None,
    )

    return result

@router.post("/bucket", summary="创建bucket")
async def create_bucket_request(
    bucket_in: BucketCreateBody,
    db: Session = Depends(get_db),
    current_user: User | None = Depends(get_auth_user_or_error),
):
    """
    创建bucket
    """
    result = await bucket_crud.create(db, obj_in=bucket_in, created_by=current_user.id if current_user else None)
    return BucketResponse(
        id=result.id,
        path=result.path,
        endpoint=result.endpoint,
        keychain_id=result.keychain_id,
        created_by=current_user.username if current_user else None,
        type=PathType.Bucket,
    )

@router.post("/bucket/batch", summary="批量创建bucket")
async def create_batch_bucket_request(
    bucket_in: List[BucketCreateBody],
    db: Session = Depends(get_db),
    current_user: User | None = Depends(get_auth_user_or_error),
):
    """
    批量创建bucket
    """
    result = await bucket_crud.create_batch(db, obj_in=bucket_in, created_by=current_user.id if current_user else None)
    return [BucketResponse(
        id=bucket.id,
        path=bucket.path,
        endpoint=bucket.endpoint,
        keychain_id=bucket.keychain_id,
        created_by=current_user.username if current_user else None,
        type=PathType.Bucket,
    ) for bucket in result]

@router.patch("/bucket/{id}", summary="更新bucket")
async def update_bucket_request(
    id: int,
    bucket_in: BucketUpdateBody,
    db: Session = Depends(get_db),
    current_user: User | None = Depends(get_auth_user_or_error),
):
    """
    更新bucket
    """
    result = await bucket_crud.update(db, id=id, obj_in=BucketUpdatePayload(
        **bucket_in.model_dump(),
        updated_by=current_user.id if current_user else None,
    ))
    return BucketResponse(
        id=result.id,
        path=result.path,
        endpoint=result.endpoint,
        created_by=current_user.username if current_user else None,
        type=PathType.Bucket,
    )

@router.get("/bucket/filter", summary="过滤bucket")
async def filter_bucket_request(
    path: str,
    db: Session = Depends(get_db),
):
    """预览s3文件"""
    if not is_s3_path(path):
        raise AppEx(
            code=ErrorCode.BUCKET_30003_INVALID_PATH,
            status_code=status.HTTP_400_BAD_REQUEST,
        )
    bucket_name, _ = split_s3_path(path)
    buckets = await bucket_crud.list_by_path(db, path=f"s3://{bucket_name}/")

    return ListResponse[BucketResponse](
        data=[BucketResponse(
            id=bucket.id,
            path=bucket.path,
            type=PathType.Bucket,
            endpoint=bucket.endpoint,
            keychain_id=bucket.keychain_id,
        ) for bucket in buckets],
        total=len(buckets)
    )

@router.get("/bucket/preview", summary="预览文件")
async def file_preview_request(
    path: str,
    mimetype: str = None,
    id: int | None = None,
    db: Session = Depends(get_db),
):
    """预览s3文件"""
    if not is_s3_path(path):
        raise AppEx(
            code=ErrorCode.BUCKET_30003_INVALID_PATH,
            status_code=status.HTTP_400_BAD_REQUEST,
        )
    result = await preview_file(
        mimetype=mimetype,
        path=path,
        id=id,
        db=db,
    )

    return result


@router.get("/bucket/download", summary="下载文件")
async def download_file_request(
    path: str,
    as_attachment: bool = True,
    id: int | None = None,
    db: Session = Depends(get_db),
):
    """
    下载指定的文件。
    """
    if not path.startswith("s3://"):
        raise AppEx(
            code=ErrorCode.BUCKET_30003_INVALID_PATH,
            status_code=status.HTTP_400_BAD_REQUEST,
        )
    
    _, s3_reader = await get_bucket(path, db, id)

    return await s3_reader.download(as_attachment=as_attachment)



@router.get("/bucket/accessible", summary="验证路径是否可访问")
async def validate_path_accessibility_request(
    path: str,
    endpoint: str,
    keychain_id: int,
    db: Session = Depends(get_db),
):
    """
    验证路径是否可访问
    """
    keychain = await keychain_crud.get(db, id=keychain_id)
    result = await validate_path_accessibility(path, endpoint, keychain.access_key_id, keychain.decrypted_secret_key_id)

    return OkResponse(data=result)


@router.get("/bucket/ping", summary="验证endpoint是否可用")
async def make_ping_request(url: str):
    """
    验证endpoint是否可用
    """

    result = ping_host(url)

    return OkResponse(data=result)


@router.get("/bucket/{id}", summary="获取bucket详情", response_model=BucketResponse)
async def get_bucket_request(
    id: int,
    db: Session = Depends(get_db),
    current_user: User | None = Depends(get_auth_user_or_error),
):
    """
    获取bucket详情
    """
    result = None

    if not settings.ENABLE_AUTH:
        result = await bucket_crud.get(db, id=id)
    else:
        result = await bucket_crud.get_by_user(db, id=id, user_id=current_user.id)

    return BucketResponse(
        id=result.id,
        path=result.path,
        endpoint=result.endpoint,
        keychain_id=result.keychain_id,
        created_by=current_user.username if current_user else None,
        type=PathType.Bucket,
    )
    

@router.delete("/bucket/{id}", summary="删除bucket")
async def delete_bucket_request(
    id: int,
    db: Session = Depends(get_db),
    current_user: User | None = Depends(get_auth_user_or_error),
):
    """
    删除bucket
    """
    await bucket_crud.delete(db, id=id)
    return OkResponse()



# TODO: 删除此接口
@router.get("/bucket/size", summary="获取路径的大小")
async def get_path_size_request():
    pass



from fastapi import Depends, Request
from jose import JWTError, jwt
from pydantic import BaseModel
from sqlalchemy.ext.asyncio import AsyncSession

from vis3.internal.client.s3_reader import S3Reader
from vis3.internal.common.db import get_db
from vis3.internal.config import settings
from vis3.internal.crud.user import user_crud
from vis3.internal.utils import timer
from vis3.internal.utils.path import split_s3_path
from vis3.internal.utils.security import decrypt_secret_key


class UserCredentials(BaseModel):
    bucket_name: str | None = None
    buckets_dict: dict | None = None
    key: str
    aws_access_key_id: str | None = None
    aws_secret_access_key: str | None = None
    region_name: str | None = None
    endpoint_url: str | None = None


def make_ak_sk(bucket_info: dict) -> tuple[str, str]:
    ak = (
        bucket_info.get("ak")
        if bucket_info.get("ak")
        else bucket_info.get("ak_sk").split("/")[0]
    )
    sk = (
        bucket_info.get("sk")
        if bucket_info.get("sk")
        else bucket_info.get("ak_sk").split("/")[1]
    )
    sk = decrypt_secret_key(sk)

    return ak, sk


async def process_request_user(request: Request, db: AsyncSession):
    """
    处理请求中的用户信息
    
    如果请求已经有用户信息，直接返回；
    否则尝试从cookie中获取令牌并解析用户
    """
    # 如果请求中已经有用户信息，直接返回
    if hasattr(request.state, "user"):
        return
    
    # 如果没有用户信息，尝试从cookie中获取令牌
    access_token = request.cookies.get("access_token")
    if not access_token:
        return
    
    try:
        # 解析令牌
        payload = jwt.decode(
            access_token, 
            settings.PASSWORD_SECRET_KEY, 
            algorithms=[settings.TOKEN_GENERATE_ALGORITHM]
        )
        user_id = payload.get("sub")
        if not user_id:
            return
        
        # 获取用户
        user = await user_crud.get(db, id=int(user_id))
        if user:
            # 将用户信息附加到请求状态
            request.state.user = user
    except JWTError:
        pass


async def get_s3_info(
    request: Request,
    db: AsyncSession = Depends(get_db),
) -> UserCredentials:
    path = request.query_params.get("path")
    bucket_name, key = split_s3_path(path)
    
    # 处理请求中的用户信息
    if settings.ENABLE_AUTH:
        await process_request_user(request, db)
        authenticated = hasattr(request.state, "user")
    else:
        # 如果未启用鉴权，视为未认证
        authenticated = False

    with timer("get_s3_info"):
        # 获取bucket信息的辅助函数
        def create_credentials(bucket_info: dict | None) -> UserCredentials:
            return UserCredentials(
                key=key,
                bucket_name=bucket_name,
                aws_access_key_id=bucket_info.get("ak") if bucket_info else None,
                aws_secret_access_key=bucket_info.get("sk") if bucket_info else None,
                region_name=bucket_info.get("region") if bucket_info else None,
                endpoint_url=bucket_info.get("endpoint") if bucket_info else None,
            )

        # 1. 检查自定义bucket
        if authenticated:
            # TODO
            _, buckets_dict = (None, {})
            if bucket_info := buckets_dict.get(bucket_name):
                ak, sk = make_ak_sk(bucket_info)
                bucket_info["ak"] = ak
                bucket_info["sk"] = sk
                return create_credentials(bucket_info)

        return None


async def get_s3_reader(
    request: Request, s3_info: UserCredentials = Depends(get_s3_info)
) -> S3Reader:
    return S3Reader(
        bucket_name=s3_info.bucket_name,
        key=s3_info.key,
        aws_access_key_id=s3_info.aws_access_key_id if s3_info.aws_access_key_id else None,
        aws_secret_access_key=s3_info.aws_secret_access_key if s3_info.aws_secret_access_key else None,
        region_name=s3_info.region_name if s3_info.region_name else None,
        endpoint_url=s3_info.endpoint_url if s3_info.endpoint_url else None,
    )



from typing import Any, Generic, List, TypeVar

from pydantic import BaseModel

T = TypeVar("T")


class ListResponse(BaseModel, Generic[T]):
    data: List[T]
    total: int


class ItemResponse(BaseModel, Generic[T]):
    data: T


class OkResponse(BaseModel):
    ok: bool = True
    message: str = "OK"
    data: Any = {}



from fastapi import APIRouter, Depends, status
from sqlalchemy.orm import Session

from vis3.internal.api.dependencies.auth import (
    get_auth_user_or_error,
)
from vis3.internal.api.v1.schema.request.keychain import (
    KeychainCreateBody,
    KeychainUpdateBody,
)
from vis3.internal.api.v1.schema.response import ListResponse
from vis3.internal.api.v1.schema.response.keychain import KeyChainResponse
from vis3.internal.common.db import get_db
from vis3.internal.common.exceptions import AppEx, ErrorCode
from vis3.internal.crud.keychain import keychain_crud
from vis3.internal.models.keychain import KeyChain
from vis3.internal.models.user import User

router = APIRouter(tags=["S3钥匙串"])

def make_keychain_response(keychain: KeyChain) -> KeyChainResponse:
    return KeyChainResponse(
        id=keychain.id,
        name=keychain.name,
        access_key_id=keychain.access_key_id,
        secret_key_id=keychain.secret_key_id,
        created_at=keychain.created_at,
        created_by=keychain.user.username if keychain.user else None,
        updated_at=keychain.updated_at,
    )


@router.get("/keychain", response_model=ListResponse[KeyChainResponse])
async def get_keychains(
    page_no: int = 1,
    page_size: int = 100,
    db: Session = Depends(get_db),
    current_user: User | None = Depends(get_auth_user_or_error),
) -> ListResponse[KeyChainResponse]:
    """
    获取当前用户的所有钥匙串
    """
    # 否则只获取当前用户的钥匙串
    keychains, total = await keychain_crud.get_multi_by_user(
        db, user_id=current_user.id, skip=(page_no - 1) * page_size, limit=page_size
    ) if current_user else await keychain_crud.get_multi(db, skip=(page_no - 1) * page_size, limit=page_size)
    return ListResponse(
        data=[
            make_keychain_response(keychain)
            for keychain in keychains
        ],
        total=total,
    )


@router.get("/keychain/all", response_model=ListResponse[KeyChainResponse])
async def get_all_keychains(
    db: Session = Depends(get_db),
    current_user: User | None = Depends(get_auth_user_or_error),
) -> ListResponse[KeyChainResponse]:
    """
    获取当前用户的所有钥匙串
    """
    # 获取当前用户的钥匙串
    keychains, total = await keychain_crud.get_all_by_user(
        db, user_id=current_user.id
    ) if current_user else await keychain_crud.get_all(db)
    return ListResponse(
        data=[
            make_keychain_response(keychain)
            for keychain in keychains
        ],
        total=total,
    )


@router.post("/keychain", response_model=KeyChainResponse, status_code=status.HTTP_201_CREATED)
async def create_keychain(
    keychain_in: KeychainCreateBody,
    db: Session = Depends(get_db),
    current_user: User | None = Depends(get_auth_user_or_error),
) -> KeyChainResponse:
    """
    创建新的钥匙串
    """
    # 创建与用户关联的钥匙串
    keychain = await keychain_crud.create_with_user(
        db, obj_in=keychain_in, user_id=current_user.id
    ) if current_user else await keychain_crud.create(db, obj_in=keychain_in)
    return make_keychain_response(keychain)


@router.get("/keychain/{keychain_id}", response_model=KeyChainResponse)
async def get_keychain(
    keychain_id: int,
    db: Session = Depends(get_db),
    current_user: User | None = Depends(get_auth_user_or_error),
) -> KeyChainResponse:
    """
    获取指定的钥匙串
    """
    keychain = await keychain_crud.get(db, id=keychain_id)
    if not keychain:
        raise AppEx(
            code=ErrorCode.KEYCHAIN_20001_KEYCHAIN_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )
    
    if current_user and keychain.created_by != current_user.id:
        raise AppEx(
            code=ErrorCode.KEYCHAIN_20003_KEYCHAIN_NOT_OWNER,
            status_code=status.HTTP_403_FORBIDDEN,
        )

    return KeyChainResponse(
        id=keychain.id,
        name=keychain.name,
        access_key_id=keychain.access_key_id,
        secret_key_id=keychain.secret_key_id,
        created_at=keychain.created_at,
        created_by=keychain.user.username if keychain.user else None,
        updated_at=keychain.updated_at,
    )


@router.patch("/keychain/{keychain_id}", response_model=KeyChainResponse)
async def update_keychain(
    keychain_id: int,
    keychain_in: KeychainUpdateBody,
    db: Session = Depends(get_db),
    current_user: User | None = Depends(get_auth_user_or_error),
) -> KeyChainResponse:
    """
    更新钥匙串
    """
    keychain = await keychain_crud.get(db, id=keychain_id)
    if not keychain:
        raise AppEx(
            code=ErrorCode.KEYCHAIN_20001_KEYCHAIN_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )
    
    # 如果启用了鉴权，检查权限
    if current_user and keychain.created_by != current_user.id:
        raise AppEx(
            code=ErrorCode.KEYCHAIN_20005_KEYCHAIN_NOT_OWNER,
            status_code=status.HTTP_403_FORBIDDEN,
        )
    
    keychain = await keychain_crud.update(db=db, id=keychain_id, obj_in=keychain_in)
    return make_keychain_response(keychain)


@router.delete("/keychain/{keychain_id}", response_model=KeyChainResponse)
async def delete_keychain(
    keychain_id: int,
    db: Session = Depends(get_db),
    current_user: User | None = Depends(get_auth_user_or_error),
) -> KeyChainResponse:
    """
    删除钥匙串（软删除）
    """
    keychain = await keychain_crud.get(db=db, id=keychain_id)
    if not keychain:
        raise AppEx(
            code=ErrorCode.KEYCHAIN_20001_KEYCHAIN_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )
    
    # 如果启用了鉴权，检查权限
    if current_user and keychain.created_by != current_user.id:
        raise AppEx(
            code=ErrorCode.KEYCHAIN_20004_KEYCHAIN_NOT_OWNER,
            status_code=status.HTTP_403_FORBIDDEN,
        )
    
    keychain = await keychain_crud.delete(db=db, id=keychain_id)
    return make_keychain_response(keychain)


from typing import List, Tuple

from fastapi import status
from sqlalchemy import func
from sqlalchemy.future import select
from sqlalchemy.orm import Session

from vis3.internal.api.v1.schema.request.bucket import (
    BucketCreateBody,
    BucketCreatePayload,
    BucketUpdatePayload,
)
from vis3.internal.common.exceptions import AppEx, ErrorCode
from vis3.internal.crud.base import BaseCrud
from vis3.internal.crud.keychain import keychain_crud
from vis3.internal.models.bucket import Bucket
from vis3.internal.schema.state import State


class BucketCRUD(BaseCrud[Bucket, BucketCreatePayload, BucketUpdatePayload]):
    async def get_by_path(self, db: Session, *, path: str, keychain_id: int | None = None) -> Bucket:
        if keychain_id:
            return db.query(self.model).filter(self.model.path.like(f"{path}%"), self.model.keychain_id == keychain_id, self.model.state == State.ENABLED).first()
        else:
            return db.query(self.model).filter(self.model.path.like(f"{path}%"), self.model.state == State.ENABLED).first()
    
    async def get_by_user_id(self, db: Session, *, user_id: int) -> Tuple[List[Bucket], int]:
        query = select(self.model).filter(self.model.created_by == user_id, self.model.state == State.ENABLED)
        count_query = select(func.count()).select_from(self.model).filter(self.model.created_by == user_id, self.model.state == State.ENABLED)
        result = db.execute(query).scalars().all()
        total = db.scalar(count_query)
        return result, total
    
    async def list_by_path(self, db: Session, *, path: str) -> List[Bucket]:
        return db.query(self.model).filter(self.model.path.like(f"{path}%"), self.model.state == State.ENABLED).all()
    
    async def create(self, db: Session, *, obj_in: BucketCreateBody, created_by: int | None = None) -> Bucket:
        keychain = await keychain_crud.get(db, id=obj_in.keychain_id)
        if not keychain:
            raise AppEx(
                code=ErrorCode.KEYCHAIN_20001_KEYCHAIN_NOT_FOUND,
                status_code=status.HTTP_404_NOT_FOUND,
            )
        
        # path + keychain_id 唯一
        if await self.get_by_path(db, path=obj_in.path, keychain_id=obj_in.keychain_id):
            raise AppEx(
                code=ErrorCode.KEYCHAIN_20002_KEYCHAIN_ALREADY_EXISTS,
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        db_obj = Bucket(
            path=obj_in.path,
            endpoint=obj_in.endpoint,
            created_by=created_by,
            keychain_id=obj_in.keychain_id,
        )
        db.add(db_obj)
        db.commit()
        db.refresh(db_obj)

        return db_obj
    
    async def create_batch(self, db: Session, *, obj_in: List[BucketCreateBody], created_by: int | None = None) -> List[Bucket]:
        db_objs = []
        
        for bucket in obj_in:
            db_obj = Bucket(
                path=bucket.path,
                endpoint=bucket.endpoint,
                created_by=created_by,
                keychain_id=bucket.keychain_id,
            )
            db.add(db_obj)
            db_objs.append(db_obj)

        db.commit()
        for db_obj in db_objs:
            db.refresh(db_obj)

        return db_objs


bucket_crud = BucketCRUD(Bucket)


import os

from appdirs import user_data_dir

_DIR_APP_NAME = "vis3"


def get_data_dir():
    data_dir = user_data_dir(appname=_DIR_APP_NAME)
    os.makedirs(data_dir, exist_ok=True)
    return data_dir



import base64
import json
import os
import re
import shutil
import tempfile
import time
import zipfile
from contextlib import contextmanager

import httpx
import orjson
import pkg_resources
from botocore.exceptions import ClientError
from bs4 import BeautifulSoup
from fastapi import status
from loguru import logger

from vis3.internal.common.exceptions import AppEx, ErrorCode
from vis3.internal.utils.path import split_s3_path


@contextmanager
def timer(operation: str):
    start = time.time()
    yield
    duration = time.time() - start
    # 毫秒
    logger.debug(f"{operation} took {duration * 1000:.2f} ms")


def is_valid_ip(ip: str) -> bool:
    pattern = re.compile(
        r"""
        ^
        (?:(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9])\.){3}
        (?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9])
        $
    """,
        re.VERBOSE,
    )
    return pattern.match(ip) is not None


# 是否需要直接读取文件内容，图片、视频、音频、pdf、epub类型直接调用预览接口
def should_not_read_as_raw(mimetype: str):
    result = (
        mimetype.startswith("image/")
        or mimetype.startswith("video/")
        or mimetype.startswith("audio/")
        or mimetype.startswith("application/pdf")
        or mimetype.startswith("application/epub")
    )

    return result


def ping_host(url: str) -> bool:
    if not url.startswith("http"):
        url = "http://" + url

    try:
        response = httpx.get(url, timeout=5)
        # 检查状态码是否在有效范围内
        return response.status_code < 500
    except httpx.RequestError:
        return False
    except Exception:
        raise AppEx(
            code=ErrorCode.PING_50000_ERROR,
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )
    
async def validate_path_accessibility(path: str, endpoint: str, ak: str, sk: str):
    from vis3.internal.client.s3_reader import S3Reader

    is_endpoint_valid = ping_host(endpoint)

    if not is_endpoint_valid:
        return False

    client = S3Reader.get_client(
        ak=ak,
        sk=sk,
        endpoint=endpoint,
        region_name=None,
    )
    bucket_name, prefix = split_s3_path(path)

    def _list_objects():
        return client.list_objects(Bucket=bucket_name, Prefix=prefix, MaxKeys=1)

    try:
        res = await S3Reader._run_in_executor(_list_objects)
        contents = res.get("Contents", [])
        logger.info(contents)
    except ClientError:
        return False

    return True


def convert_mobi_stream_to_html(mobi_content: bytes) -> str:
    import mobi

    """
	将mobi文件内容流转换为包含内嵌资源的HTML
	"""
    # 使用BytesIO创建一个内存中的文件对象
    with tempfile.NamedTemporaryFile(delete=False, suffix=".mobi") as temp_file:
        temp_file.write(mobi_content)
        temp_file.flush()
        temp_file_path = temp_file.name

    # 使用 mobi 库将 MOBI 文件解包到临时目录
    mobi_temp_dir, _ = mobi.extract(temp_file_path)

    # 查找解包后的 EPUB 或 HTML 文件
    epub_file = None
    html_files = []
    image_files = {}

    # 收集所有图片文件
    for root, _, files in os.walk(mobi_temp_dir):
        for file in files:
            if file.endswith(".epub"):
                epub_file = os.path.join(root, file)
            elif file.endswith(".html"):
                html_files.append(os.path.join(root, file))
            elif file.lower().endswith((".jpg", ".jpeg", ".png", ".gif")):
                # 存储图片路径，以文件名为键
                image_files[file] = os.path.join(root, file)

    # 如果存在 EPUB 文件，则进一步解压 EPUB 文件
    if epub_file:
        with zipfile.ZipFile(epub_file, "r") as zip_ref:
            zip_ref.extractall(mobi_temp_dir)

        # 再次查找解压后的所有文件
        for root, _, files in os.walk(mobi_temp_dir):
            for file in files:
                if file.endswith(".html"):
                    html_files.append(os.path.join(root, file))
                elif file.lower().endswith((".jpg", ".jpeg", ".png", ".gif")):
                    # 存储图片路径，以文件名为键
                    image_files[file] = os.path.join(root, file)

    # 合并所有 HTML 文件内容
    html_content = ""
    for html_file in html_files:
        with open(html_file, "r", encoding="utf-8", errors="ignore") as f:
            content = f.read()
            soup = BeautifulSoup(content, "html.parser")

            # 处理图片标签，将图片转为base64嵌入
            for img_tag in soup.find_all("img"):
                src = img_tag.get("src")
                if src:
                    # 提取文件名
                    img_filename = os.path.basename(src)
                    if img_filename in image_files:
                        # 读取图片并转换为base64
                        try:
                            with open(image_files[img_filename], "rb") as img_file:
                                img_data = img_file.read()
                                img_type = os.path.splitext(img_filename)[1][1:].lower()
                                if img_type == "jpg":
                                    img_type = "jpeg"
                                base64_data = base64.b64encode(img_data).decode("utf-8")
                                # 更新图片源为base64
                                img_tag["src"] = (
                                    f"data:image/{img_type};base64,{base64_data}"
                                )
                        except Exception as e:
                            logger.error(f"处理图片时出错: {e}")
                            img_tag.replace_with(
                                soup.new_tag(
                                    "div",
                                    attrs={"class": "image-placeholder"},
                                    string="图片加载失败",
                                )
                            )
                    else:
                        # 如果找不到图片文件，显示占位符
                        img_tag.replace_with(
                            soup.new_tag(
                                "div",
                                attrs={"class": "image-placeholder"},
                                string="图片未找到",
                            )
                        )

            html_content += str(soup)

    # 删除临时文件
    shutil.rmtree(mobi_temp_dir)
    os.remove(temp_file_path)

    return html_content


def convert_epub_stream_to_html(epub_content: bytes) -> str:
    import ebooklib
    from ebooklib import epub

    book = None

    with tempfile.NamedTemporaryFile(delete=False, suffix=".epub") as temp_file:
        temp_file.write(epub_content)
        temp_file.flush()

        book = epub.read_epub(temp_file.name)

    html_content = ""
    # 收集所有图片资源
    image_resources = {}

    # 提取所有图片资源
    for item in book.get_items_of_type(ebooklib.ITEM_IMAGE):
        # 获取图片ID和内容
        image_id = item.get_id()
        image_content = item.get_content()
        image_type = item.media_type.split("/")[1]
        if image_type == "jpg":
            image_type = "jpeg"
        # 将图片内容转为base64
        base64_data = base64.b64encode(image_content).decode("utf-8")
        image_resources[image_id] = f"data:image/{image_type};base64,{base64_data}"

    for item in book.get_items_of_type(ebooklib.ITEM_DOCUMENT):
        content = item.get_content().decode("utf-8", errors="ignore")
        soup = BeautifulSoup(content, "html.parser")

        # 处理图片标签
        for img_tag in soup.find_all("img"):
            src = img_tag.get("src")
            if src:
                # EPUB中图片src通常使用相对路径或ID引用
                img_id = src.split("/")[-1]
                # 尝试查找匹配的图片资源
                for resource_id, resource_data in image_resources.items():
                    if img_id in resource_id or resource_id in img_id:
                        img_tag["src"] = resource_data
                        break
                else:
                    # 如果找不到图片资源
                    img_tag.replace_with(
                        soup.new_tag(
                            "div",
                            attrs={"class": "image-placeholder"},
                            string="图片未找到",
                        )
                    )

        # 删除a标签里的href
        for a_tag in soup.find_all("a"):
            a_tag.name = "span"
            if "href" in a_tag.attrs:
                del a_tag.attrs["href"]

        html_content += f"<section>{str(soup.body) if soup.body else ''}</section>"

    full_html = f"""
	<!DOCTYPE html>
	<html lang="en">
	<head>
		<meta charset="UTF-8">
		<meta name="viewport" content="width=device-width, initial-scale=1.0">
		<title>EPUB Preview</title>
		<style>
			body {{ font-family: Arial, sans-serif; line-height: 1.6; padding: 1rem 5rem; background-color: #f2f2f2; }}
			img {{ max-width: 100%; height: auto; }}
			section {{ background-color: #fff; padding: 2rem 4rem; margin-top: 2rem; }}
			section:first-child {{ margin-top: 0; }}
			.image-placeholder {{ padding: 1rem; background-color: #eee; text-align: center; color: #666; }}
		</style>
	</head>
	<body>
		{html_content}
	</body>
	</html>
	"""

    # 删除临时文件
    os.remove(temp_file.name)

    return full_html


def json_dumps(d: dict, **kwargs) -> str:
    if not kwargs and orjson:
        try:
            return orjson.dumps(d).decode("utf-8")
        except Exception:
            pass
    return json.dumps(d, ensure_ascii=False, **kwargs)

def update_sys_config(config: dict):
    frontend_public = os.path.join(
        pkg_resources.resource_filename('vis3.internal', 'statics'),
    )
    os.makedirs(frontend_public, exist_ok=True)

    with open(os.path.join(frontend_public, "sys-config.js"), "w", encoding="utf-8") as f:
        f.write(f"(function() {{ window.__CONFIG__ = {json_dumps(config)}; }})();")



from pydantic import BaseModel, Field


class UserBase(BaseModel):
    username: str = Field(..., min_length=3, max_length=50)


class Token(BaseModel):
    access_token: str
    token_type: str


class TokenPayload(BaseModel):
    sub: str = None
    exp: int = None 


from typing import Tuple
from urllib.parse import parse_qsl, quote, urlparse

from fastapi import status
from fastapi.responses import HTMLResponse, StreamingResponse
from loguru import logger
from sqlalchemy.orm import Session

from vis3.internal.api.v1.schema.response import ItemResponse, ListResponse
from vis3.internal.api.v1.schema.response.bucket import BucketResponse, PathType
from vis3.internal.client.s3_reader import S3Reader
from vis3.internal.common.exceptions import AppEx, ErrorCode
from vis3.internal.crud.bucket import bucket_crud
from vis3.internal.models.bucket import Bucket
from vis3.internal.utils import (
    convert_epub_stream_to_html,
    convert_mobi_stream_to_html,
    should_not_read_as_raw,
    timer,
)
from vis3.internal.utils.path import is_s3_path, split_s3_path


async def get_bucket(path: str, db: Session, id: int | None = None) -> Tuple[Bucket, S3Reader]:
    bucket_name, key = split_s3_path(path)
    
    if id:
        bucket = await bucket_crud.get(db, id=id)
    else:
        buckets = await bucket_crud.list_by_path(db, path=f"s3://{bucket_name}/")

        if len(buckets) > 0:
            bucket = buckets[0]
        else:
            raise AppEx(
                code=ErrorCode.BUCKET_30001_OBJECT_NOT_FOUND,
                status_code=status.HTTP_404_NOT_FOUND,
            )

    if not bucket:
        raise AppEx(
            code=ErrorCode.BUCKET_30001_OBJECT_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )
    
    s3_reader = S3Reader(
        bucket=bucket,
        bucket_name=bucket_name,
        key=key,
        endpoint_url=bucket.endpoint,
        access_key_id=bucket.keychain.access_key_id,
        secret_access_key=bucket.keychain.decrypted_secret_key_id,
    )

    return bucket, s3_reader


async def get_s3_directories(
    s3_reader: S3Reader, page_no, page_size
):
    list_of_gn = s3_reader.list_objects(
        recursive=False,
        page_no=page_no,
        page_size=page_size,
    )

    result = []

    async for gn in list_of_gn:
        _path = gn[0] if isinstance(gn, tuple) else gn
        details = gn[1] if isinstance(gn, tuple) else {}

        target_type = PathType.Directory if _path.endswith("/") else PathType.File
        # 基础信息
        owner = details.get("Owner")
        display_name = owner.get("DisplayName") if owner else ""
        owner_id = owner.get("ID") if owner else ""
        last_modified = details.get("LastModified")
        size = details.get("Size") or details.get("ContentLength")

        result.append(
            BucketResponse(
                type=target_type,
                path=_path,
                owner=f"{display_name}/{owner_id}" if owner else None,
                last_modified=last_modified,
                size=size,
                id=s3_reader.bucket.id,
            )
        )

    return result


async def get_file(parsed_path: str, query_dict: dict, s3_reader: S3Reader):
    with timer("get_file"):
        file_header_info = await s3_reader.head_object()
        size = file_header_info.get("ContentLength", 0) if file_header_info else 0

        if size == 0:
            return BucketResponse(
                type=PathType.File,
                id=s3_reader.bucket.id,
                owner=await s3_reader.get_object_owner(),
                size=size,
                path=s3_reader.path,
                last_modified=file_header_info.get("LastModified")
                if file_header_info
                else None,
            )

        mimetype = await s3_reader.mime_type()

        if mimetype and should_not_read_as_raw(mimetype):
            return BucketResponse(
                type=PathType.File,
                id=s3_reader.bucket.id,
                owner=await s3_reader.get_object_owner(),
                size=size,
                mimetype=mimetype,
                path=s3_reader.path,
                last_modified=file_header_info.get("LastModified"),
            )

        if file_header_info is None:
            raise AppEx(
                code=ErrorCode.BUCKET_30001_OBJECT_NOT_FOUND,
                status_code=status.HTTP_404_NOT_FOUND,
            )

        owner = await s3_reader.get_object_owner()
        chunk_size = min(size, 1 << 20)

        if not chunk_size:
            return BucketResponse(
                type=PathType.File,
                id=s3_reader.bucket.id,
                owner=owner,
                size=size,
                mimetype=mimetype,
                content="",
                path=s3_reader.path,
                last_modified=file_header_info.get("LastModified"),
            )

        request_byte_start = query_dict.get("bytes", "").split(",")[0]
        request_byte_start = (
            int(request_byte_start)
            if request_byte_start or request_byte_start != ""
            else 0
        )

        if request_byte_start and request_byte_start >= size:
            raise AppEx(
                code=ErrorCode.BUCKET_30002_OUT_OF_RANGE,
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        # 文件
        if parsed_path.endswith(".jsonl") or parsed_path.endswith(".jsonl.gz") or parsed_path.endswith(".warc.gz"):
            row = await s3_reader.read_row(start=request_byte_start)

            return BucketResponse(
                type=PathType.File,
                id=s3_reader.bucket.id,
                owner=owner,
                size=size,
                mimetype=mimetype,
                last_modified=file_header_info.get("LastModified"),
                content=row.value,
                path=row.loc,
            )

        rest_size = size - request_byte_start

        if rest_size <= 0:
            raise AppEx(
                code=ErrorCode.BUCKET_30002_OUT_OF_RANGE,
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        # 读取文本
        logger.info(f"read s3 object: {s3_reader.key_without_query}, rest size: {rest_size}")
        # 读取非jsonl的文本内容，依次读取1mb的内容
        read_size = min(rest_size, 1 << 20)
        byte_range = f"{request_byte_start},{read_size}"
        content = ""

        async for chunk, _ in s3_reader.read_by_range(
            start_byte=request_byte_start,
            end_byte=request_byte_start + read_size,
        ):
            content += chunk.decode("utf-8", errors="ignore")

        return BucketResponse(
            type=PathType.File,
            id=s3_reader.bucket.id,
            owner=owner,
            size=size,
            mimetype=mimetype,
            last_modified=file_header_info.get("LastModified"),
            content=content,
            path=f"s3://{s3_reader.bucket_name}/{s3_reader.key_without_query}?bytes={byte_range}",
        )


async def get_buckets_or_objects(
    path: str,
    page_no: int,
    page_size: int,
    db: Session,
    id: int | None = None,
    user_id: int | None = None,
):
    """获取bucket或目录或文件
    """
    result = None
    # 获取所有bucket列表
    if path is None or path == "" or path == "/":
        buckets, total = await bucket_crud.get_by_user_id(db, user_id=user_id) if user_id else await bucket_crud.get_multi(db)
        result = [
            BucketResponse(
                id=bucket.id,
                keychain_id=bucket.keychain_id,
                keychain_name=bucket.keychain.name,
                type=PathType.Bucket
                if bucket.path.endswith("/")
                else PathType.File,
                path=bucket.path,
                owner=bucket.user.username if bucket.user else None,
                endpoint=bucket.endpoint,
                last_modified=None,
                size=None,
            )
            for bucket in buckets
        ]

        return ListResponse[BucketResponse](data=result, total=total)
    
    _, s3_reader = await get_bucket(path, db, id)


    path_without_query, _, query = path.partition("?")
    s3_path = quote(path_without_query, safe=":/")
    parsed_url = urlparse(s3_path)
    parsed_path = parsed_url.path

    if not is_s3_path(path):
        raise AppEx(
            code=ErrorCode.BUCKET_30003_INVALID_PATH,
            status_code=status.HTTP_400_BAD_REQUEST,
        )
    
    # 目录
    if parsed_path.endswith("/"):
        with timer("get s3 dirs"):
            result = await get_s3_directories(
                s3_reader=s3_reader,
                page_no=page_no,
                page_size=page_size,
            )
        return ListResponse[BucketResponse](data=result, total=len(result))

    # 文件
    result = await get_file(
        parsed_path=parsed_path,
        query_dict=dict(parse_qsl(query)),
        s3_reader=s3_reader,
    )

    return ItemResponse[BucketResponse](data=result)


async def preview_file(
    path: str,
    db: Session,
    mimetype: str = None,
    id: int | None = None,
):
    """
    预览文件内容，支持多种文件格式。

    Args:
        mimetype: 文件的 MIME 类型
        s3_reader: S3Reader 实例

    Returns:
        StreamingResponse: 文件流响应
        HTMLResponse: 对于特殊格式（如 EPUB、MOBI）的 HTML 响应

    Raises:
        AppEx: 当文件不存在或大小超出限制时
    """
    _, s3_reader = await get_bucket(path, db, id)
    # 获取文件信息
    file_header_info = await s3_reader.head_object()
    if file_header_info is None:
        raise AppEx(
            code=ErrorCode.BUCKET_30001_OBJECT_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    file_size = file_header_info.get("ContentLength")
    max_file_size = 40 << 20  # 40 MB

    if file_size > max_file_size:
        raise AppEx(
            code=ErrorCode.BUCKET_30002_OUT_OF_RANGE,
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    # 获取文件类型
    if not mimetype:
        mimetype = await s3_reader.mime_type()
        if not mimetype:
            raise AppEx(
                code=ErrorCode.BUCKET_30005_DATA_IS_EMPTY,
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )

    # 处理特殊格式文件
    if mimetype in ("application/x-mobipocket-ebook", "application/epub+zip"):
        # 使用流式读取处理大文件
        chunks = []
        async for chunk, _ in s3_reader.read_by_range(start_byte=0):
            chunks.append(chunk)
            if len(chunks) > max_file_size:  # 限制内存使用
                raise AppEx(
                    code=ErrorCode.BUCKET_30002_OUT_OF_RANGE,
                    status_code=status.HTTP_400_BAD_REQUEST,
                )

        full_content = b"".join(chunks)
        if mimetype == "application/x-mobipocket-ebook":
            html_content = convert_mobi_stream_to_html(full_content)
        else:
            html_content = convert_epub_stream_to_html(full_content)
        return HTMLResponse(content=html_content, media_type="text/html")

    # 对于其他文件类型，使用流式响应
    async def content_generator():
        # 使用分块读取，避免一次性加载大文件
        chunk_size = 1024 * 1024  # 1MB chunks
        offset = 0

        try:
            while offset < file_size:
                end = min(offset + chunk_size, file_size)
                async for chunk, _ in s3_reader.read_by_range(
                    start_byte=offset, end_byte=end
                ):
                    yield chunk
                offset = end
        except Exception as e:
            logger.error(f"文件流处理错误: {str(e)}")
            raise

    response = StreamingResponse(
        content_generator(),
        media_type=mimetype,
        headers={
            "Accept-Ranges": "bytes",
        },
    )

    return response


from fastapi import status
from sqlalchemy.future import select
from sqlalchemy.orm import Session

from vis3.internal.api.v1.schema.request.keychain import (
    KeychainCreateBody,
    KeychainUpdateBody,
)
from vis3.internal.common.exceptions import AppEx, ErrorCode
from vis3.internal.crud.base import BaseCrud
from vis3.internal.models.keychain import KeyChain
from vis3.internal.schema.state import State
from vis3.internal.utils.security import encrypt_secret_key


class KeyChainCRUD(BaseCrud[KeyChain, KeychainCreateBody, KeychainUpdateBody]):
    async def _get_by_access_key_id(self, db: Session, *, access_key_id: str, user_id: int | None = None) -> KeyChain:
        """
        获取钥匙串
        """
        result = db.execute(select(KeyChain).filter(KeyChain.access_key_id == access_key_id, KeyChain.state == State.ENABLED, KeyChain.created_by == user_id))
        return result.scalars().first()

    async def get(self, db: Session, *, id: int) -> KeyChain:
        """
        获取钥匙串
        """
        result = db.execute(select(KeyChain).filter(KeyChain.id == id, KeyChain.state == State.ENABLED))
        return result.scalars().first()

    async def create_with_user(
        self, db: Session, *, obj_in: KeychainCreateBody, user_id: int
    ) -> KeyChain:
        """
        创建新钥匙串，关联到用户
        """
        if await self._get_by_access_key_id(db, access_key_id=obj_in.access_key_id, user_id=user_id):
            raise AppEx(
                code=ErrorCode.KEYCHAIN_20002_KEYCHAIN_ALREADY_EXISTS,
                status_code=status.HTTP_400_BAD_REQUEST,
            )
        
        db_obj = KeyChain(
            name=obj_in.name,
            access_key_id=obj_in.access_key_id,
            # 加密 sk
            secret_key_id=encrypt_secret_key(obj_in.secret_key_id),
            created_by=user_id,
        )
        db.add(db_obj)
        db.commit()
        db.refresh(db_obj)
        return db_obj
        
    async def create(self, db: Session, *, obj_in: KeychainCreateBody) -> KeyChain:
        """
        创建新钥匙串，不关联用户（用于未启用鉴权时）
        """
        if await self._get_by_access_key_id(db, access_key_id=obj_in.access_key_id, user_id=None):
            raise AppEx(
                code=ErrorCode.KEYCHAIN_20002_KEYCHAIN_ALREADY_EXISTS,
                status_code=status.HTTP_400_BAD_REQUEST,
            )
        
        db_obj = KeyChain(
            name=obj_in.name,
            access_key_id=obj_in.access_key_id,
            # 加密 sk
            secret_key_id=encrypt_secret_key(obj_in.secret_key_id),
        )
        db.add(db_obj)
        db.commit()
        db.refresh(db_obj)
        return db_obj


keychain_crud = KeyChainCRUD(KeyChain)


from fastapi import Depends, Request, status
from fastapi.security import OAuth2PasswordBearer
from jose import JWTError, jwt
from loguru import logger
from sqlalchemy.orm import Session

from vis3.internal.common.db import get_db
from vis3.internal.common.exceptions import AppEx, ErrorCode
from vis3.internal.config import settings
from vis3.internal.crud.user import user_crud
from vis3.internal.models.user import User
from vis3.internal.schema.user import TokenPayload

# 设置token_url，但不自动抛出错误
oauth2_scheme = OAuth2PasswordBearer(tokenUrl=f"{settings.API_V1_STR}/auth/login", auto_error=False)


async def get_token_from_request(request: Request, token: str | None = Depends(oauth2_scheme)) -> str | None:
    """
    从请求中获取token，优先从Authorization头中获取，其次从cookie中获取
    """
    if token:
        return token
    
    # 如果Authorization头中没有token，尝试从cookie中获取
    access_token = request.cookies.get("access_token")
    if access_token:
        return access_token
    
    if settings.ENABLE_AUTH:
        raise AppEx(
            code=ErrorCode.AUTH_10001_NOT_AUTHENTICATED,
            status_code=status.HTTP_401_UNAUTHORIZED,
        )
    
    return None


async def get_current_user(
    request: Request,
    db: Session = Depends(get_db), 
    token: str | None = Depends(get_token_from_request)
) -> User:
    """
    获取当前用户
    """
    logger.info(f"\(^o^)/~ get_current_user: {token}")
    
    try:
        payload = jwt.decode(
            token, settings.PASSWORD_SECRET_KEY, algorithms=[settings.TOKEN_GENERATE_ALGORITHM]
        )
        user_id: str = payload.get("sub")
        token_data = TokenPayload(sub=user_id)
    except JWTError as e:
        logger.error(f"JWT解码错误: {str(e)}")
        raise AppEx(
            code=ErrorCode.AUTH_10001_NOT_AUTHENTICATED,
            status_code=status.HTTP_401_UNAUTHORIZED,
        )
    
    user = await user_crud.get(db, id=int(token_data.sub))
    if user is None:
        logger.warning(f"找不到ID为 {token_data.sub} 的用户")
        raise AppEx(
            code=ErrorCode.AUTH_10001_NOT_AUTHENTICATED,
            status_code=status.HTTP_401_UNAUTHORIZED,
        )
    return user


async def get_auth_user_or_error(
    request: Request,
    db: Session = Depends(get_db), 
    token: str | None = Depends(get_token_from_request)
) -> User | None:
    """
    根据配置决定是否需要验证用户
    
    当ENABLE_AUTH为False时，返回None（不进行鉴权）
    当ENABLE_AUTH为True时，验证失败则抛出异常
    """
    if not settings.ENABLE_AUTH:
        return None
    
    # 当启用鉴权但未提供token时抛出异常
    if not token and settings.ENABLE_AUTH:
        raise AppEx(
            code=ErrorCode.AUTH_10001_NOT_AUTHENTICATED,
            status_code=status.HTTP_401_UNAUTHORIZED,
        )
    
    return await get_current_user(request, db, token) 


import os
import traceback
from enum import Enum
from typing import Any

import pkg_resources
from botocore.exceptions import (
    ClientError,
    EndpointConnectionError,
    ParamValidationError,
)
from fastapi import FastAPI, HTTPException, Request, status
from fastapi.exceptions import RequestValidationError
from fastapi.responses import FileResponse, JSONResponse
from loguru import logger
from starlette.exceptions import HTTPException as StarletteHTTPException

COMMON = 00000
AUTH = 10000
KEYCHAIN = 20000
BUCKET = 30000
S3_CLIENT = 40000
PING = 50000
SERVER = 60000
TRANSLATE_SERVICE_ERROR = 70000


class ErrorCode(Enum):
    # common
    CODE_00000_SQL_ERROR = (
        COMMON,
        "SQL Error",
    )
    CODE_00001_NO_PERMISSION = (
        COMMON,
        "Permission Denied",
    )
    CODE_00002_VALIDATION_ERROR = (
        COMMON + 2,
        "Request Validation Error",
    )
    CODE_00003_CLIENT_ERROR = (
        COMMON + 3,
        "Invalid Request",
    )
    AUTH_10001_NOT_AUTHENTICATED = (AUTH + 1, "Not Authenticated")
    AUTH_10002_INVALID_USERNAME = (AUTH + 2, "Invalid Username")
    AUTH_10003_INVALID_PASSWORD = (AUTH + 3, "Invalid Password")
    BUCKET_30001_OBJECT_NOT_FOUND = (BUCKET + 1, "Object Not Found, Please Check the Path")
    BUCKET_30002_OUT_OF_RANGE = (BUCKET + 2, "Request Size Out of Range, Maximum Size is 40 Mb")
    BUCKET_30003_INVALID_PATH = (BUCKET + 3, "Invalid S3 Path, Must Start with s3://")
    BUCKET_30004_PATH_IS_EMPTY = (BUCKET + 4, "S3 Path Cannot Be Empty")
    BUCKET_30005_DATA_IS_EMPTY = (BUCKET + 5, "Data is Empty")
    BUCKET_30006_CONFIG_FILE_NOT_FOUND = (BUCKET + 6, "S3 Config File Not Found")
    BUCKET_30007_DUPLICATED_BUCKETS = (BUCKET + 7, "Duplicate Bucket Names Found")
    KEYCHAIN_20001_KEYCHAIN_NOT_FOUND = (KEYCHAIN + 1, "Keychain Not Found")
    KEYCHAIN_20002_KEYCHAIN_ALREADY_EXISTS = (KEYCHAIN + 2, "Keychain Already Exists")
    KEYCHAIN_20003_KEYCHAIN_NOT_OWNER = (KEYCHAIN + 3, "No Permission to Access This Keychain")
    KEYCHAIN_20004_KEYCHAIN_NOT_OWNER = (KEYCHAIN + 4, "No Permission to Delete This Keychain")
    KEYCHAIN_20005_KEYCHAIN_NOT_OWNER = (KEYCHAIN + 5, "No Permission to Update This Keychain")
    KEYCHAIN_20006_KEYCHAIN_NOT_OWNER = (KEYCHAIN + 6, "No Permission to Access This Keychain")
    KEYCHAIN_20007_KEYCHAIN_NOT_OWNER = (KEYCHAIN + 7, "No Permission to Delete This Keychain")
    S3_CLIENT_40000_ERROR = (S3_CLIENT, "S3 Client Error")
    S3_CLIENT_40001_ACCESS_DENIED = (S3_CLIENT + 1, "Access Denied")
    S3_CLIENT_40002_NO_SUCH_BUCKET = (S3_CLIENT + 2, "Requested Path Does Not Exist")
    S3_CLIENT_40003_NOT_FOUND = (S3_CLIENT + 3, "Bucket Configuration Not Found")
    S3_CLIENT_40004_UNKNOWN_ERROR = (S3_CLIENT + 4, "Unknown Error")
    PING_50000_ERROR = (PING, "Verification Failed")
    SERVER_60000_ERROR = (SERVER, "Server Error")
    SERVER_60001_READ_ERROR = (SERVER + 1, "Failed to Read File")


class AppEx(HTTPException):
    def __init__(
        self, code: ErrorCode, status_code=status.HTTP_400_BAD_REQUEST, detail: str = ""
    ) -> None:
        self.msg = code.value[1]
        self.code = code.value[0]
        self.detail = detail
        self.status_code = status_code


async def app_exception_handler(request: Request, exp: AppEx):
    logger.error(exp)
    return JSONResponse(
        status_code=exp.status_code,
        content={
            "msg": exp.msg,
            "err_code": exp.code,
            "detail": exp.detail,
        },
    )


async def http_exception_handler(request: Request, exp: StarletteHTTPException):
    logger.error(exp)

    print(traceback.format_exc())
    logger.error(traceback.format_exc())

    if (
        isinstance(exp, StarletteHTTPException)
        and exp.status_code == status.HTTP_404_NOT_FOUND
        and not request.url.path.startswith("/api")
    ):
        return FileResponse(
            os.path.join(
                pkg_resources.resource_filename('vis3.internal', 'statics'),
                'index.html'
                ),
            status_code=200,
            headers={'Content-Type': 'text/html', 'Cache-Control': 'no-cache', 'Connection': 'keep-alive'}
        )
    elif (
        isinstance(exp, StarletteHTTPException)
        and exp.status_code == status.HTTP_403_FORBIDDEN
    ):
        return JSONResponse(
            status_code=exp.status_code,
            content={
                "err_code": ErrorCode.CODE_10001_NOT_AUTHENTICATED.value[0],
                "msg": exp.detail
                if str(exp.detail)
                else ErrorCode.CODE_10001_NOT_AUTHENTICATED.value[1],
            },
        )

    detail = None

    try:
        detail = exp.detail
    except Exception:
        detail = str(exp)

    return JSONResponse(
        status_code=hasattr(exp, "status_code") and exp.status_code or 500,
        content={
            "err_code": ErrorCode.CODE_00003_CLIENT_ERROR.value[0],
            "detail": detail,
            "msg": ErrorCode.CODE_00003_CLIENT_ERROR.value[1],
        },
    )


async def validation_exception_handler(request: Request, exp: Any):
    logger.error(exp)

    return JSONResponse(
        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
        content={
            "msg": str(exp),
            "err_code": ErrorCode.CODE_00002_VALIDATION_ERROR.value[0],
        },
    )


async def sql_exception_handler(request: Request, exp: Any):
    logger.error(exp)
    return JSONResponse(
        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        content={
            "msg": str(exp),
            "err_code": ErrorCode.CODE_00000_SQL_ERROR.value[0],
        },
    )


async def boto_exception_handler(request: Request, exp: ClientError):
    logger.error(exp)
    if isinstance(exp, ClientError):
        if exp.response.get("Error", {}).get("Code") == "AccessDenied":
            return JSONResponse(
                status_code=status.HTTP_403_FORBIDDEN,
                content={
                    "msg": str(exp),
                    "err_code": ErrorCode.S3_CLIENT_40001_ACCESS_DENIED.value[0],
                },
            )
        elif exp.response.get("Error", {}).get("Code") == "NoSuchBucket":
            return JSONResponse(
                status_code=status.HTTP_404_NOT_FOUND,
                content={
                    "msg": str(exp),
                    "err_code": ErrorCode.S3_CLIENT_40002_NO_SUCH_BUCKET.value[0],
                },
            )

    if isinstance(exp, ParamValidationError):
        return JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST,
            content={
                "msg": str(exp),
                "err_code": ErrorCode.S3_CLIENT_40000_ERROR.value[0],
            },
        )
    else:
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            content={
                "msg": str(exp),
                "err_code": ErrorCode.S3_CLIENT_40000_ERROR.value[0],
            },
        )


async def handle_unicode_decode_error(request: Request, exp: UnicodeDecodeError):
    logger.error(exp)
    return JSONResponse(
        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        content={
            "msg": str(exp),
            "err_code": ErrorCode.SERVER_60001_READ_ERROR.value[0],
        },
    )


async def handle_server_error(request: Request, exp: Exception):
    logger.error(exp)
    if isinstance(exp, HTTPException):
        return http_exception_handler(request, exp)

    return JSONResponse(
        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        content={
            "msg": str(exp),
            "err_code": ErrorCode.SERVER_60000_ERROR.value[0],
        },
    )


def add_exception_handler(app: FastAPI):
    app.add_exception_handler(AppEx, app_exception_handler)
    app.add_exception_handler(StarletteHTTPException, http_exception_handler)
    app.add_exception_handler(HTTPException, http_exception_handler)
    app.add_exception_handler(ClientError, boto_exception_handler)
    app.add_exception_handler(ParamValidationError, boto_exception_handler)
    app.add_exception_handler(RequestValidationError, validation_exception_handler)
    app.add_exception_handler(UnicodeDecodeError, handle_unicode_decode_error)
    app.add_exception_handler(EndpointConnectionError, boto_exception_handler)
    app.add_exception_handler(Exception, http_exception_handler)



from datetime import datetime, timedelta, timezone

from fastapi import APIRouter, Depends, Response, status
from sqlalchemy.orm import Session

from vis3.internal.api.dependencies.auth import get_auth_user_or_error
from vis3.internal.api.v1.schema.request.user import UserCreate, UserLogin
from vis3.internal.api.v1.schema.response.user import UserResponse
from vis3.internal.common.db import get_db
from vis3.internal.common.exceptions import AppEx, ErrorCode
from vis3.internal.config import settings
from vis3.internal.crud.user import user_crud
from vis3.internal.models.user import User
from vis3.internal.schema.state import AuthState
from vis3.internal.schema.user import Token
from vis3.internal.utils.security import create_access_token

router = APIRouter(prefix="/auth", tags=["认证"])


@router.post("/register", response_model=UserResponse)
async def register(
    user_in: UserCreate, db: Session = Depends(get_db)
) -> UserResponse:
    """
    用户注册
    """
    user = await user_crud.get_by_username(db, username=user_in.username)
    if user:
        raise AppEx(
            code=ErrorCode.AUTH_10004_USERNAME_ALREADY_EXISTS,
            status_code=status.HTTP_400_BAD_REQUEST,
        )
    user = await user_crud.create(db, obj_in=user_in)
    return user


@router.post("/login", response_model=Token)
async def login(
    response: Response,
    user_in: UserLogin,
    db: Session = Depends(get_db),
):
    """
    用户登录，获取访问令牌
    """
    result = await user_crud.authenticate(
        db, username=user_in.username, password=user_in.password
    )
    if result == AuthState.USERNAME_ERROR:
        raise AppEx(
            code=ErrorCode.AUTH_10002_INVALID_USERNAME,
            status_code=status.HTTP_400_BAD_REQUEST,
        )
    elif result == AuthState.PASSWORD_ERROR:
        raise AppEx(
            code=ErrorCode.AUTH_10003_INVALID_PASSWORD,
            status_code=status.HTTP_400_BAD_REQUEST,
        )
    else:
        user = result
    
    access_token = create_access_token(subject=user.id)

    # 设置cookie
    response.set_cookie(
        key="access_token", 
        value=access_token, 
        httponly=True, 
        secure=settings.SCHEME == "https", 
        samesite="lax",
        expires=datetime.now(timezone.utc) + timedelta(minutes=settings.TOKEN_ACCESS_EXPIRE_MINUTES),
    )
    
    return {"access_token": access_token, "token_type": settings.TOKEN_TYPE}

@router.get("/me", response_model=UserResponse)
async def get_current_user(
    current_user: User = Depends(get_auth_user_or_error),
    db: Session = Depends(get_db),
):
    """
    获取当前用户
    """
    return UserResponse(
        id=current_user.id,
        username=current_user.username,
    )

@router.get("/logout", status_code=status.HTTP_204_NO_CONTENT)
async def logout(response: Response):
    """
    用户登出，删除cookie
    """
    response.delete_cookie(key="access_token")

    return {"message": "登出成功"}


import asyncio
import codecs
import io
import json
import urllib.parse
import zlib
from typing import AsyncIterator, Optional, Tuple, Union

import boto3
import httpx
from botocore.client import Config
from botocore.exceptions import ClientError, NoCredentialsError
from fastapi import HTTPException, status
from fastapi.responses import StreamingResponse
from loguru import logger

from vis3.internal.common.exceptions import AppEx, ErrorCode
from vis3.internal.models.bucket import Bucket
from vis3.internal.schema import JsonRow
from vis3.internal.utils import json_dumps, timer
from vis3.internal.utils.path import extract_bytes_range


class FakeRedis:
    async def get(self, key: str) -> str:
        return ""

    async def set(self, key: str, value: str):
        pass


redis_client = FakeRedis()

MAX_END = 1 * 1024 * 1024

def _is_valid_charset(charset: str):
    try:
        codecs.lookup(charset)
        return True
    except LookupError:
        return False


def _try_decode(body_bytes: bytes, http_charset: Union[str, None]):
    import cchardet

    tried_charsets = set()
    # 1. try decode with `http_charset`.
    if http_charset and _is_valid_charset(http_charset):
        try:
            http_charset = http_charset.lower()
            tried_charsets.add(http_charset)
            return body_bytes.decode(http_charset), http_charset
        except:
            pass
    # 2. try decode with utf-8.
    try:
        tried_charsets.add("utf-8")
        return body_bytes.decode("utf-8"), "utf-8"
    except:
        pass
    # 3. try detect charset and decode.
    charset = cchardet.detect(body_bytes).get("encoding")
    if charset:
        charset = charset.lower()
        if charset in ["gbk", "gb2312"]:
            charset = "gb18030"
        if charset not in tried_charsets:
            try:
                return body_bytes.decode(charset), charset
            except:
                pass
    # 4. gave up.
    return "", ""



class S3Reader:
    # 添加文件签名映射
    FILE_SIGNATURES = {
        # ZIP 文件
        b'PK\x03\x04': 'application/zip',
        # PDF 文件
        b'%PDF': 'application/pdf',
        # GIF 文件
        b'GIF87a': 'image/gif',
        b'GIF89a': 'image/gif',
        # JPEG 文件
        b'\xFF\xD8\xFF': 'image/jpeg',
        # PNG 文件
        b'\x89PNG\r\n\x1a\n': 'image/png',
        # GZIP 文件
        b'\x1F\x8B\x08': 'application/gzip',
        # XML 文件 (UTF-8)
        b'<?xml': 'application/xml',
        # UTF-8 文本文件的 BOM
        b'\xEF\xBB\xBF': 'text/plain',
        # UTF-16 LE 文本文件的 BOM
        b'\xFF\xFE': 'text/plain',
        # UTF-16 BE 文本文件的 BOM
        b'\xFE\xFF': 'text/plain',
    }

    # 保持现有的 MIME_TYPES 映射
    MIME_TYPES = {
        '.txt': 'text/plain',
        '.html': 'text/html',
        '.htm': 'text/html',
        '.json': 'application/json',
        '.jsonl': 'application/json',
        '.csv': 'text/csv',
        '.xml': 'application/xml',
        '.pdf': 'application/pdf',
        '.jpg': 'image/jpeg',
        '.jpeg': 'image/jpeg',
        '.png': 'image/png',
        '.gif': 'image/gif',
        '.zip': 'application/zip',
        '.gz': 'application/gzip',
        '.warc': 'application/warc',
        '.doc': 'application/msword',
        '.docx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
        '.xls': 'application/vnd.ms-excel',
        '.xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
        '.js': 'application/javascript',
        '.css': 'text/css',
    }

    def __init__(
        self,
        key: str,
        bucket: Bucket,
        access_key_id: str,
        secret_access_key: str,
        region_name: str = "us-east-1",
        bucket_name: str | None = None,
        endpoint_url: str | None = None,
    ):
        self.bucket_name = bucket_name
        self.key = key
        self.bucket = bucket
        self.path = f"s3://{bucket_name}/{key}"
        self.key_without_query = key.split("?")[0]
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.region_name = region_name
        self.endpoint_url = endpoint_url
        self.is_compressed = self.key_without_query.endswith(".gz")

        if self.access_key_id and self.secret_access_key:
            self.client = S3Reader.get_client(self.access_key_id, self.secret_access_key, self.endpoint_url, self.region_name)

    @staticmethod
    async def _run_in_executor(func, *args, **kwargs):
        loop = asyncio.get_event_loop()

        # 创建一个包装函数来处理关键字参数
        def wrapper():
            return func(*args, **kwargs)

        return await loop.run_in_executor(None, wrapper)

    def _get_range_header(self, start: int, length: int | None = None):
        max_end = start + MAX_END
        if length is None or length == 0:
            # 如果 offset 为空，读取从 start 开始的 1MB 数据
            # 这个大小足够读取一行，同时避免读取过多数据
            range_header = f"bytes={start}-{max_end}"
        else:
            range_header = f"bytes={start}-{start + length}"

        return range_header
    
    @staticmethod
    def get_client(ak: str, sk: str, endpoint: str, region_name: str):
        try:
            return boto3.client(
                "s3",
                aws_access_key_id=ak,
                aws_secret_access_key=sk,
                region_name=region_name,
                endpoint_url=endpoint,
                config=Config(s3={"addressing_style": "virtual"})
            )
        except Exception:
            # TODO: 错误类型
            return boto3.client(
                "s3",
                aws_access_key_id=ak,
                aws_secret_access_key=sk,
                endpoint_url=endpoint,
                config=Config(s3={"addressing_style": "path"}, retries={"max_attempts": 8}),
            )

    async def head_object(self):
        """
        获取 S3 对象的头部信息。

        Args:
            key: S3 对象的键

        Returns:
            dict: 对象的头部信息

        Raises:
            ValueError: 当 bucket 不存在时
            PermissionError: 当没有访问权限时
            ClientError: 其他 S3 客户端错误
        """
        try:
            with timer("head_object"):
                return await S3Reader._run_in_executor(
                    self.client.head_object,
                    Bucket=self.bucket_name,
                    Key=self.key_without_query,
                )
        except ClientError as e:
            error_code = e.response["Error"]["Code"]
            if error_code == "NoSuchBucket":
                raise AppEx(
                    code=ErrorCode.S3_CLIENT_40002_NO_SUCH_BUCKET,
                    status_code=status.HTTP_404_NOT_FOUND,
                    detail=f"Bucket {self.bucket_name} does not exist",
                )
            elif error_code == "AccessDenied":
                raise AppEx(
                    code=ErrorCode.S3_CLIENT_40001_ACCESS_DENIED,
                    status_code=status.HTTP_403_FORBIDDEN,
                    detail=f"Access denied to bucket {self.bucket_name}",
                )
            elif error_code == "404":
                raise AppEx(
                    code=ErrorCode.S3_CLIENT_40003_NOT_FOUND,
                    status_code=status.HTTP_404_NOT_FOUND,
                    detail=f"Object {self.key_without_query} does not exist in bucket {self.bucket_name}",
                )
            else:
                raise
        except Exception as e:
            if isinstance(e, NoCredentialsError):
                raise AppEx(
                    code=ErrorCode.S3_CLIENT_40001_ACCESS_DENIED,
                    status_code=status.HTTP_403_FORBIDDEN,
                )
            raise

    async def _detect_mime_by_content(self) -> str:
        """
        通过文件内容的签名来检测 MIME 类型
        """
        try:
            # 获取文件的前32个字节用于类型检测
            response = await S3Reader._run_in_executor(
                self.client.get_object,
                Bucket=self.bucket_name,
                Key=self.key_without_query,
                Range="bytes=0-31",
            )
            content = await S3Reader._run_in_executor(response["Body"].read)

            # 检查文件签名
            for signature, mime_type in self.FILE_SIGNATURES.items():
                if content.startswith(signature):
                    return mime_type

            # 检查是否是文本文件
            # 检查前32个字节是否都是可打印字符或常见控制字符
            is_text = all(byte in b' \t\n\r' or 32 <= byte <= 126 for byte in content)
            if is_text:
                # 如果内容以 { [ " 开头，可能是 JSON
                if content.lstrip().startswith(b'{') or content.lstrip().startswith(b'['):
                    return 'application/json'
                return 'text/plain'

            return 'application/octet-stream'

        except Exception as e:
            logger.error(f"Error detecting MIME type from content: {e}")
            return 'application/octet-stream'

    async def mime_type(self) -> str:
        """
        获取文件的 MIME 类型。优先使用文件扩展名，如果没有扩展名则通过内容检测。

        Returns:
            str: 文件的 MIME 类型
        """
        try:
            # 处理特殊情况：.warc.gz
            if self.key_without_query.endswith('.warc.gz'):
                return 'application/warc+gzip'

            # 尝试从文件扩展名获取 MIME 类型
            if '.' in self.key_without_query:
                file_ext = '.' + self.key_without_query.split('.')[-1].lower()
                mime_type = self.MIME_TYPES.get(file_ext)
                if mime_type:
                    return mime_type

            # 如果没有扩展名或扩展名未知，通过内容检测
            return await self._detect_mime_by_content()

        except Exception as e:
            logger.error(f"Error determining MIME type: {e}")
            return 'application/octet-stream'

    async def list_objects(
        self,
        recursive=False,
        limit=0,
        page_size=100,
        page_no=1,
    ):
        """
        分页获取 S3 对象列表

        Args:
            recursive: 是否递归获取子目录
            limit: 最大返回数量，0 表示不限制
            page_size: 每页大小
            page_no: 当前页码，从 1 开始

        Yields:
            Tuple[str, dict, str]: (s3_url, 对象详情, 类型)
        """
        marker = None
        item_yielded = 0
        current_page = 1

        while True:
            operation_parameters = {
                "Bucket": self.bucket_name,
                "Prefix": self.key_without_query,
                "MaxKeys": page_size,
                "Delimiter": "/" if not recursive else None,
            }
            if marker:
                operation_parameters["Marker"] = marker

            try:
                result = await S3Reader._run_in_executor(
                    self.client.list_objects, **operation_parameters
                )
            except ClientError as e:
                error_code = e.response["Error"]["Code"]
                if error_code == "NoSuchBucket":
                    raise ValueError(f"Bucket {self.bucket_name} does not exist")
                elif error_code == "AccessDenied":
                    raise PermissionError(f"Access denied to bucket {self.bucket_name}")
                else:
                    raise

            contents = result.get("Contents", [])
            common_prefixes = result.get("CommonPrefixes", [])
            next_marker = result.get("NextMarker")

            if current_page == page_no:
                for content in contents:
                    if not content["Key"].endswith("/"):
                        yield (
                            f"s3://{self.bucket_name}/{content['Key']}",
                            content,
                            "file",
                        )
                        item_yielded += 1
                        if limit > 0 and item_yielded >= limit:
                            return
                        if item_yielded == page_size:
                            return

                for _prefix in common_prefixes:
                    yield (
                        f"s3://{self.bucket_name}/{_prefix['Prefix']}",
                        _prefix,
                        "directory",
                    )
                    item_yielded += 1
                    if limit > 0 and item_yielded >= limit:
                        return
                    if item_yielded == page_size:
                        return

            if not next_marker or item_yielded == page_size:
                break

            marker = next_marker
            current_page += 1

    async def read_warc_gz(
        self,
        start: int | None = None,
        length: int | None = None,
    ) -> JsonRow:
        """
        读取 WARC.GZ 文件记录，并以指定JSON格式返回。

        Args:
            start: 起始字节位置
            offset: 结束字节位置

        Returns:
            JsonRow: 包含 WARC 记录内容的 JsonRow 对象，使用以下格式:
            {
                "record_id": "track_id",
                "url": "记录的URL",
                "status": "响应状态码",
                "response_header": "响应头部信息",
                "date": "记录日期",
                "content_length": "记录大小",
                "html": "实际内容",
                "remark": "原始头部信息"
            }
        """

        try:
            from fastwarc.warc import ArchiveIterator, WarcRecordType

            response = await S3Reader._run_in_executor(
                self.client.get_object,
                Bucket=self.bucket_name,
                Key=self.key_without_query,
                Range=self._get_range_header(start=start, length=length),
                RequestPayer="requester",
            )
            stream = response["Body"]
            # 读取 StreamingBody 的内容
            content = await stream.read()
            file_obj = io.BytesIO(content)

            def process_warc():
                result = None
                record_length = 0
                next_start = start

                try:
                    # 使用 fastwarc 库解析 WARC 文件
                    archive_iterator = ArchiveIterator(file_obj)
                    record_found = False

                    for record in archive_iterator:
                        if result:
                            pre_stream_pos = result.get("remark", {}).get(
                                "stream_pos", 0
                            )
                            next_start = pre_stream_pos + next_start
                            record_length = record.stream_pos - pre_stream_pos
                            break

                        if record.record_type == WarcRecordType.response:
                            # 获取 HTTP 响应内容
                            http_headers = {}
                            warc_headers = {}

                            # 处理 WARC 头部
                            for name, value in record.headers.items():
                                if name.startswith("WARC-"):
                                    warc_headers[name] = value

                            # 获取 HTTP 头部和内容
                            http_status = None

                            if record.http_headers:
                                http_status = record.http_headers.status_code
                                for name, value in record.http_headers.items():
                                    http_headers[name] = value

                            if http_status is None or http_status >= 400:
                                continue

                            # 获取内容
                            try:
                                content_bytes = record.reader.read()
                            except:
                                result["content_length"] = -1
                                content_bytes = None

                            # 尝试解码内容
                            html_content = ""
                            charset = None
                            content_charset = ""
                            content_length = len(content_bytes) if content_bytes else -1

                            # 从 Content-Type 头部提取字符集
                            content_type = http_headers.get("Content-Type", "")
                            if "charset=" in content_type:
                                charset = (
                                    content_type.split("charset=")[1]
                                    .split(";")[0]
                                    .strip()
                                )

                            # 尝试解码内容
                            if content_bytes:
                                html_content, content_charset = _try_decode(
                                    content_bytes, charset
                                )

                            # 构建结果
                            result = {
                                "track_id": str(record.record_id).split(":")[-1][:36],
                                "url": record.headers.get("WARC-Target-URI", ""),
                                "status": http_status,
                                "response_header": http_headers,
                                "date": record.record_date.timestamp(),
                                "content_length": content_length,
                                "html": html_content,
                                "content_charset": content_charset,
                                "remark": {
                                    "warc_headers": warc_headers,
                                    "stream_pos": record.stream_pos,
                                },
                            }

                            record_found = True

                    if not record_found:
                        result = {"error": "No WARC response record found"}
                except Exception as e:
                    result = {"error": f"Error reading record: {str(e)}"}

                return result, next_start, record_length

            # 在事件循环中执行处理（因为处理可能是CPU密集型的）
            loop = asyncio.get_event_loop()
            result, start, length = await loop.run_in_executor(None, process_warc)

            return JsonRow(
                value=json_dumps(result),
                loc=self._make_location(start, length),
                metadata={"content_length": result.get("content_length", 0)},
            )
        except Exception as e:
            raise AppEx(
                code=ErrorCode.S3_CLIENT_40004_UNKNOWN_ERROR,
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"Error reading record: {str(e)}",
            )

    async def read_by_range(
        self, start_byte: int, end_byte: int | None = None
    ) -> AsyncIterator[Tuple[Union[str, bytes, dict], int]]:
        """
        按字节范围读取文件内容，支持文本和压缩文件。

        Args:
            start_byte: 起始字节位置
            end_byte: 结束字节位置，如果为 None 则读取到文件末尾

        Yields:
            Tuple[Union[str, bytes, dict], int]: (内容块, 偏移量)
            - 对于文本文件：返回 (文本内容, 偏移量)
            - 对于压缩文件：返回 (解压后的内容, 偏移量)
            - 对于 JSONL 文件：返回 (解析后的 JSON 对象, 偏移量)
        """

        # 创建包装函数正确处理get_object
        def get_object():
            return self.client.get_object(
                Bucket=self.bucket_name,
                Key=self.key_without_query,
                Range=f"bytes={start_byte}-{end_byte}"
                if end_byte
                else f"bytes={start_byte}-",
                RequestPayer="requester",
            )

        response = await S3Reader._run_in_executor(get_object)
        stream = response["Body"]
        current_byte = start_byte

        if not self.is_compressed:
            # 对于非压缩文件，直接获取内容并分块返回
            def read_chunks():
                return stream.read()

            content = await S3Reader._run_in_executor(read_chunks)

            # 按块返回内容
            chunk_size = 8192  # 8KB 块大小
            for i in range(0, len(content), chunk_size):
                chunk = content[i : i + chunk_size]
                yield chunk, current_byte
                current_byte += len(chunk)
        else:
            # 对于压缩文件，先读取全部内容再解压处理
            def read_and_decompress():
                data = stream.read()
                decompressor = zlib.decompressobj(32 + zlib.MAX_WBITS)
                decompressed_data = decompressor.decompress(data)
                decompressed_data += decompressor.flush()
                return decompressed_data

            decompressed_content = await S3Reader._run_in_executor(read_and_decompress)

            if self.key_without_query.endswith(".jsonl.gz"):
                # 对于 JSONL 文件，解析每一行
                lines = decompressed_content.splitlines()
                for line in lines:
                    if line:
                        try:
                            yield json.loads(line.decode("utf-8")), current_byte
                        except json.JSONDecodeError:
                            yield line, current_byte
                        current_byte += len(line) + 1  # +1 for newline
            else:
                # 按块返回解压后的内容
                chunk_size = 8192
                for i in range(0, len(decompressed_content), chunk_size):
                    chunk = decompressed_content[i : i + chunk_size]
                    yield chunk, current_byte
                    current_byte += len(chunk)

    async def get_object_owner(self):
        try:
            async for _, details, _ in self.list_objects(limit=1):
                owner = details.get("Owner")

            return f"{owner.get('DisplayName')}/{owner.get('ID')}" if owner else None

        except Exception as e:
            print(e)
            return None

    def _make_location(self, start: int, offset: Optional[int] = None):
        return f"s3://{self.bucket_name}/{self.key_without_query}?bytes={start},{offset}"

    async def read_s3_row_with_cache(self, start: int, length: int | None = None):
        """
        读取S3行，并缓存结果，Need redis support
        """
        row = None
        cache_key = f"s3_svc:s3://{self.bucket_name}/{self.key}"
        cached_result = redis_client.get(cache_key)

        if cached_result:
            cached_row = json.loads(cached_result)
            row = JsonRow(value=cached_row.get("row"), loc=cached_row.get("path"))
        else:
            row = await self.read_row(start=start, length=length)
            redis_client.set(
                cache_key, json.dumps({"row": row.value, "path": row.loc}), ex=120
            )

        # cache next row
        asyncio.create_task(self.cache_s3_next_row(path=row.loc))

        return row

    async def cache_s3_next_row(self, path: str):
        path, offset, length = extract_bytes_range(path)
        next_offset = offset + length

        next_row_cache_key = f"s3_svc:{path}?bytes={next_offset},0"

        cached_result = redis_client.get(next_row_cache_key)

        if cached_result:
            return

        # 如果是jsonl，最大获取2mb
        length = 2 << 20 if self.key_without_query.endswith(".jsonl") else None

        next_row = await self.read_row(start=next_offset, length=length)
        redis_client.set(
            next_row_cache_key,
            json.dumps({"row": next_row.value, "path": next_row.loc}),
            ex=120,
        )

    async def read_row(
        self,
        start: int,
        length: int | None = None,
    ) -> JsonRow:
        """
        根据字节范围读取一行内容。

        Args:
            start: 起始字节位置
            offset: 结束字节位置，如果为 None 则读取从 start 开始的完整一行

        Returns:
            JsonRow: (行内容, 偏移量, 行号)
        """
        if self.is_compressed:
            return await self.read_gz_row(start=start, length=length)

        # 变成异步
        MAX_TOTAL_SIZE = 10 * 1024 * 1024  # 10MB
        NEXT_READ_SIZE = 1 * 1024 * 1024  # 1MB

        current_start = start
        total_size = 0
        buffer = bytearray()

        while total_size < MAX_TOTAL_SIZE:
            # 计算本次读取的大小
            read_size = NEXT_READ_SIZE

            # 变成异步
            def get_object_sync():
                return self.client.get_object(
                    Bucket=self.bucket_name,
                    Key=self.key_without_query,
                    Range=self._get_range_header(start=current_start, length=read_size),
                    RequestPayer="requester",
                )

            response = await self._run_in_executor(get_object_sync)
            stream = response["Body"]

            # 读取数据
            chunk = await self._run_in_executor(stream.read)
            if not chunk:
                break

            buffer.extend(chunk)
            total_size += len(chunk)

            # 查找换行符
            newline_pos = buffer.find(b"\n")
            if newline_pos != -1:
                # 找到换行符，提取内容
                line = buffer[:newline_pos]
                new_start = start
                new_len = len(line) + 1

                # 处理行内容
                try:
                    decoded_line = line.decode("utf-8")
                except UnicodeDecodeError:
                    try:
                        decoded_line = line.decode("latin1")
                    except UnicodeDecodeError:
                        decoded_line = str(line)

                return JsonRow(
                    value=decoded_line,
                    loc=self._make_location(new_start, new_len),
                    offset=new_len,
                    size=len(line),
                )

            current_start += len(chunk)

        # 如果达到最大限制还没找到换行符，返回所有内容
        try:
            decoded_line = buffer.decode("utf-8")
        except UnicodeDecodeError:
            try:
                decoded_line = buffer.decode("latin1")
            except UnicodeDecodeError:
                decoded_line = str(buffer)

        return JsonRow(
            value=decoded_line,
            loc=self._make_location(start, len(buffer)),
            offset=len(buffer),
            size=len(buffer),
        )

    async def read_gz_row(self, start: int, length: int | None = None) -> JsonRow:
        """
        根据字节范围读取压缩文件中的一行内容。

        Args:
            start: 起始字节位置
            length: 结束字节位置，如果为 None 则读取从 start 开始的完整一行

        Returns:
            JsonRow: (行内容, 偏移量, 行号)
        """
        if self.key_without_query.endswith(".warc.gz"):
            return await self.read_warc_gz(start=start, length=length)

        try:
            # 从文件开头读取数据
            response = await S3Reader._run_in_executor(
                self.client.get_object,
                Bucket=self.bucket_name,
                Key=self.key_without_query,
                Range=self._get_range_header(start, length),
            )
            stream = response["Body"]

            # 使用warcio的BufferedReader来处理gz文件
            from warcio.bufferedreaders import BufferedReader

            # 创建一个字节流对象
            buff_reader = BufferedReader(stream, decomp_type="gzip")

            # 读取解压后的内容
            line = None
            original_length = 0

            while True:
                line = buff_reader.readline()

                if line:
                    original_length = stream.tell() - buff_reader.rem_length()
                    break
                elif buff_reader.read_next_member():
                    continue
                else:
                    break

            try:
                # 尝试解析为JSON
                decoded_line = line.decode("utf-8").rstrip("\r\n")
                # 验证是否为有效的JSON
                json.loads(decoded_line)
            except (UnicodeDecodeError, json.JSONDecodeError):
                try:
                    decoded_line = line.decode("latin1")
                except Exception:
                    decoded_line = str(line)

            return JsonRow(
                value=decoded_line,
                loc=self._make_location(start, original_length),
                offset=original_length,
            )

        except Exception as e:
            logger.error(f"Error reading gz file: {e}")
            return JsonRow(value="", loc=self._make_location(start, 0), offset=0)

    async def get_s3_presigned_url(self, as_attachment=True) -> str:
        params = {"Bucket": self.bucket_name, "Key": self.key_without_query}
        if as_attachment:
            filename = self.key_without_query.split("/")[-1]
            params["ResponseContentDisposition"] = f'attachment; filename="{filename}"'

        # 创建包装函数正确处理所有参数
        def generate_url():
            return self.client.generate_presigned_url("get_object", Params=params)

        return await S3Reader._run_in_executor(generate_url)

    async def download(self, as_attachment=True) -> StreamingResponse:
        try:
            presigned_url = await self.get_s3_presigned_url(as_attachment=as_attachment)
            file_header_info = await self.head_object()
            chunk_size = 1024 * 1024

            async def stream_file():
                async with httpx.AsyncClient() as http_client:
                    try:
                        async with http_client.stream("GET", presigned_url) as response:
                            if response.status_code != 200:
                                raise AppEx(
                                    code=ErrorCode.S3_CLIENT_40004_UNKNOWN_ERROR,
                                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                                    detail=f"Failed to download file: {response.text}",
                                )

                            async for chunk in response.aiter_bytes(
                                chunk_size=chunk_size
                            ):
                                yield chunk
                    except httpx.RequestError as e:
                        raise AppEx(
                            code=ErrorCode.S3_CLIENT_40004_UNKNOWN_ERROR,
                            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                            detail=f"Failed to stream file: {str(e)}",
                        )
                    except Exception as e:
                        raise AppEx(
                            code=ErrorCode.S3_CLIENT_40004_UNKNOWN_ERROR,
                            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                            detail=f"Unexpected error while streaming file: {str(e)}",
                        )

            headers = {
                "Content-Disposition": f'attachment; filename="{urllib.parse.quote(self.key_without_query)}"',
                "Content-Type": file_header_info.get(
                    "ContentType", "application/octet-stream"
                ),
            }

            return StreamingResponse(
                stream_file(),
                headers=headers,
                media_type=file_header_info.get(
                    "ContentType", "application/octet-stream"
                ),
            )
        except Exception as e:
            if isinstance(e, HTTPException):
                raise
            raise AppEx(
                code=ErrorCode.S3_CLIENT_40004_UNKNOWN_ERROR,
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"Failed to prepare download: {str(e)}",
            )



from typing import Any, Dict, Generic, List, Optional, Tuple, Type, TypeVar, Union

from pydantic import BaseModel
from sqlalchemy import func
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from vis3.internal.schema.state import State

T = TypeVar("T", bound=BaseModel)
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)


class BaseCrud(Generic[T, CreateSchemaType, UpdateSchemaType]):
    def __init__(self, model: Type[T]):
        self.model = model

    async def get(self, db: AsyncSession, id: int) -> T | None:
        """
        获取单个对象
        """
        result = db.execute(select(self.model).filter(self.model.id == id))
        return result.scalars().first()
    
    async def get_by_user(self, db: AsyncSession, id: int, user_id: int) -> T | None:
        """
        获取单个对象
        """
        result = db.execute(select(self.model).filter(self.model.id == id, self.model.created_by == user_id))
        return result.scalars().first()
    
    async def get_all(self, db: AsyncSession) -> Tuple[List[T], int]:
        """
        获取所有对象
        """
        query = select(self.model).filter(self.model.state == State.ENABLED)
        count_query = select(func.count()).select_from(self.model).filter(self.model.state == State.ENABLED)
        if hasattr(self.model, "created_at"):
            query = query.order_by(self.model.created_at.desc())
        result = db.execute(query).scalars().all()
        total = db.scalar(count_query)
        return result, total
    
    async def get_all_by_user(self, db: AsyncSession, user_id: int) -> Tuple[List[T], int]:
        """
        获取所有对象
        """
        query = select(self.model).filter(self.model.created_by == user_id, self.model.state == State.ENABLED)
        count_query = select(func.count()).select_from(self.model).filter(self.model.created_by == user_id, self.model.state == State.ENABLED)
        if hasattr(self.model, "created_at"):
            query = query.order_by(self.model.created_at.desc())
        result = db.execute(query).scalars().all()
        total = db.scalar(count_query)
        return result, total

    async def get_multi(
        self, db: AsyncSession, *, skip: int = 0, limit: int = 100
    ) -> Tuple[List[T], int]:
        """
        获取多个对象
        """
        query = select(self.model).filter(self.model.state == State.ENABLED).offset(skip).limit(limit)
        count_query = select(func.count()).select_from(self.model).filter(self.model.state == State.ENABLED)
        if hasattr(self.model, "created_at"):
            query = query.order_by(self.model.created_at.desc())
        result = db.execute(query).scalars().all()
        total = db.scalar(count_query)
        return result, total
    

    async def get_multi_by_user(
        self, db: AsyncSession, *, user_id: int, skip: int = 0, limit: int = 100
    ) -> Tuple[List[T], int]:
        """
        获取多个对象
        """
        query = select(self.model).filter(self.model.created_by == user_id, self.model.state == State.ENABLED).offset(skip).limit(limit)
        count_query = select(func.count()).select_from(self.model).filter(self.model.created_by == user_id, self.model.state == State.ENABLED)
        if hasattr(self.model, "created_at"):
            query = query.order_by(self.model.created_at.desc())
        result = db.execute(query).scalars().all()
        total = db.scalar(count_query)
        return result, total

    async def create(self, db: AsyncSession, *, obj_in: CreateSchemaType, auto_commit: bool = True) -> T:
        """
        创建对象
        
        参数:
            db: 数据库会话
            obj_in: 创建对象的输入模型
            auto_commit: 是否自动提交事务，默认为True。设置为False时可以在外部管理事务。
        """
        obj_dict = obj_in.model_dump()
        db_obj = self.model(**obj_dict)
        
        db.add(db_obj)
        if auto_commit:
            db.commit()
            db.refresh(db_obj)
        
        return db_obj

    async def update(
        self, db: AsyncSession, *, id: int, obj_in: Union[UpdateSchemaType, Dict[str, Any]],
        auto_commit: bool | None = True
    ) -> T | None:
        """
        更新对象
        
        参数:
            db: 数据库会话
            id: 对象ID
            obj_in: 更新对象的输入模型或字典
            auto_commit: 是否自动提交事务，默认为True。设置为False时可以在外部管理事务。
        """
        db_obj = await self.get(db=db, id=id)
        if db_obj is None:
            return None

        update_data = (
            obj_in if isinstance(obj_in, dict) else obj_in.model_dump(exclude_unset=True)
        )

        for field in update_data:
            if hasattr(db_obj, field) and update_data[field] is not None:
                setattr(db_obj, field, update_data[field])

        db.add(db_obj)
        if auto_commit:
            db.commit()
            db.refresh(db_obj)
        
        return db_obj

    async def delete(self, db: AsyncSession, *, id: Any, auto_commit: bool = True) -> Optional[T]:
        """
        删除对象（软删除）
        
        将对象的state更新为disabled，而不是硬删除
        
        参数:
            db: 数据库会话
            id: 对象ID
            auto_commit: 是否自动提交事务，默认为True。设置为False时可以在外部管理事务。
        """
        db_obj = await self.get(db=db, id=id)
        if db_obj is None:
            return None

        # 软删除：将state设置为disabled
        if hasattr(db_obj, "state"):
            setattr(db_obj, "state", State.DISABLED)
            db.add(db_obj)
            if auto_commit:
                db.commit()
                db.refresh(db_obj)
        
        return db_obj



from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware


class ContentTypeMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        response = await call_next(request)
        
        # set content-type for javascript files
        if request.url.path.endswith(".js"):
            response.headers["content-type"] = "application/javascript"

        if request.url.path.endswith("sys-config.js"):
            # 禁用缓存
            response.headers["Cache-Control"] = "no-cache"
            response.headers["Pragma"] = "no-cache"
            response.headers["Expires"] = "0"
            
        return response


from datetime import datetime
from enum import StrEnum

from pydantic import BaseModel


class PathType(StrEnum):
    Bucket = "bucket"
    Directory = "directory"
    File = "file"


class BucketResponse(BaseModel):
    # common
    id: int
    path: str
    type: PathType
    # bucket only
    endpoint: str | None = None
    created_by: str | None = None
    # file only
    mimetype: str | None = None
    size: int | None = None
    content: str | None = None
    last_modified: datetime | None = None
    keychain_id: int | None = None
    keychain_name: str | None = None


from datetime import datetime

from sqlalchemy import Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import relationship
from vis3.internal.common.db import Base
from vis3.internal.schema.state import State
from vis3.internal.utils.security import decrypt_secret_key


class KeyChain(Base):
    __tablename__ = "keychain"
    id = Column(Integer, primary_key=True, autoincrement=True, index=True)
    name = Column(String(255))
    access_key_id = Column(String(255))
    secret_key_id = Column(String(255))
    created_at = Column(DateTime(timezone=True), default=datetime.now, comment="Time a user was created")
    updated_at = Column(DateTime(timezone=True), default=datetime.now, comment="Last time a user was updated")
    created_by = Column(Integer, ForeignKey(column="user.id"), index=True)
    state = Column(String(255), default=State.ENABLED)

    # relation
    buckets = relationship("Bucket", back_populates="keychain")
    user = relationship("User", foreign_keys=[created_by])

    # 获取解密后的 secret_key_id
    @property
    def decrypted_secret_key_id(self):
        return decrypt_secret_key(self.secret_key_id)



from datetime import datetime

from sqlalchemy import Column, DateTime, Integer, String
from sqlalchemy.orm import relationship
from vis3.internal.common.db import Base


class User(Base):
    __tablename__ = "user"

    id = Column(Integer, primary_key=True, autoincrement=True, index=True)
    username = Column(String(50), unique=True, index=True)
    hashed_password = Column(String(255))
    created_at = Column(
        DateTime(timezone=True), default=datetime.now, comment="Time a user was created"
    )
    updated_at = Column(
        DateTime(timezone=True), default=datetime.now, comment="Last time a user was updated"
    )

    # relation
    keychains = relationship("KeyChain", back_populates="user")
    buckets = relationship("Bucket", back_populates="user")